/*
   File: Move.prg

   Replacement for CA-Clipper dbf movement functions
   Each function attempts some method of verifying the order/index
      value to attempt detection of corrupt order/index files
   While the detection ability is not foolproof it is a huge
      advance for us since we had none at all
   #include 'move.ch' in ALL programs EXCEPT move.prg, recompile
      all CA-Clipper source and link in move.prg
   Tom Claffy - August 10, 1995
*/

#include 'error.ch'
#include 'set.ch'

/*
  Replacement function for DbSeek. A successful dbseek is only half of
  the process: dbseek returns .t. when the search value is located in
  the order file and subsequently positions the dbf file to the offset
  returned. This replacement function expands the data in the dbf to
  verify that the dbf and order file are in sync.
*/

Function MvSeek( xValue, lSoftSeek )
LOCAL cExpr
LOCAL lReturn := .f.
If( lSoftseek = NIL, lSoftSeek := .f. , )  // default value
DbSkip(0)                                  // refresh network indexes
Dbseek( xValue, lSoftSeek )
If (lReturn := ( !Eof() ))
   If lSoftSeek
     /*
        NOTE - Uncomment the following line to maintain full
        compatibility with CA-Clipper. I prefer to have a return
        value for softseek of .not. EOF() while CA-Clipper returns .f.
        unless found()
     */
     // lReturn := .f.
   Else
       /*
          expand the value of the index key in the record and compare
          the expanded value to the passed parameter
       */
      cExpr := &( IndexKey(0) )
      If( Valtype( xValue ) = "C",  cExpr :=  Left( cExpr ,Len( xValue )), )
      If( !( xValue == cExpr ), MvError(), )
   Endif
Endif
RETURN lReturn

/*
  The following replacement functions for the movement functions
  are slightly more difficult to test. The possibility of legitimate
  duplicate records negates the use of record numbers as a test so
  I have chosen to stick with the expansion of the key. Remember
  that the order file is a separate file containing a duplication
  of the order key data and that is the test objective.
*/

Function MvGoTo( nRecord )
DbGoTo( nRecord )
MvVerify()
RETURN NIL

Function MvGoTop
DbGoTop()
MvVerify()
RETURN NIL

Function MvGoBottom
DbGoBottom()
MvVerify()
RETURN NIL

Function MvSkip( nRecords )
If( nRecords = NIL, nRecords := 1 , )
DbSkip( nRecords )
MvVerify()
RETURN NIL

STATIC Function MvVerify
LOCAL cExpr := IndexKey(0)
LOCAL xValue, yValue
LOCAL nRecordNumber := Recno()
If (!EOF() .and. ;
    !DELETED() .and. ;
    !Empty(cExpr) .and. ;
    DELETED() = SET(_SET_DELETED))
   /*
      store the expanded value and try to find it again
   */
   DbSkip(0)                           // refresh network indexes
   xValue := ( &cExpr )
   DbSeek( xValue  )
   /*
      compare the stored value to the expanded value of the
      record found - could be a different record if duplicates
      are allowed so we are just testing the first occurrence
      of this key
   */
   yValue := ( &cExpr )
   If( !( xValue == yValue ), MvError(), )
   /*
      Duplicate records may have resulted in the incorrect positioning
      of the record pointer during the seek, so restore it
   */
   DbGoTo( nRecordNumber )
Endif
RETURN NIL

/*
   Generate an error to be emailed to us by the errorsys.prg
   with a detailed report of the error conditions
   We get this over the WAN. You can take appropriate action
   for your situation.
*/

STATIC Function MvError
LOCAL oErrorBlock := ErrorBlock()
LOCAL oError := ErrorNew()
oError:Description := "Index Corruption Detected - REINDEXING REQUIRED"
oError:FileName := Alias()
oError:Severity := ES_CATASTROPHIC
Eval( oErrorBlock , oError )
RETURN NIL

/*
   generic conversion of differing data types to character
*/

STATIC Function XTOC( xChar )
LOCAL aConversions := { { "C" , { |x| x }                } , ;
                        { "D" , { |x| Dtoc( x ) }        } , ;
                        { "L" , { |x| If( x, "T","F" ) } } , ;
                        { "N" , { |x| Ltrim( Str( x )) } }     }
LOCAL cType := VALTYPE( xChar )
LOCAL nPosition := Ascan( aConversions, { |x| x[1] = cType } )
RETURN If( nPosition # 0, Eval( aConversions[ nPosition, 2 ], xChar ), "" )

